﻿using System;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;

namespace TestVoiceNET
{
    class SoftCodec : CopyWav
    {
        [DllImport("BoODLL.dll")]
        private static extern void Init_G723_1();

        [DllImport("BoODLL.dll")]
        private static extern unsafe void Coder_G723_1(short* DataBuff, byte* Vout);

        [DllImport("BoODLL.dll")]
        private static extern unsafe void Decoder_G723_1(short* DataBuff, byte* Vinp);

        public SoftCodec(System.Windows.Forms.Control aControl) :
            base(aControl)
        {
            m_Form = (Form1)aControl;

            m_LKDSPro.Application("TestVoiceNet");
            m_LKDSPro.ReceiveVoice += LKDSPro_ReceiveVoice;
            m_LKDSPro.ReceiveMsg += LKDSPro_ReceiveMsg;

            byte body = 0;
            m_LKDSPro.SendPack(1, 0, 0, ref body);  // !!! subscribe to messages from LU index 1 - "SendPack(uint NumLB, ..."
            //m_LKDSPro.SendPack(2, 0, 0, ref body);  // !!! subscribe to messages from LU index 2
            //m_LKDSPro.SendPack(3, 0, 0, ref body);  // !!! subscribe to messages from LU index 3
            // ...
        }

        public void Start(int aIndexLU, int aPoint)
        {
            m_IndexLU = aIndexLU;
            m_Point = aPoint;

            m_Id = Guid.NewGuid();

            Init_G723_1();

            base.Start();
        }

        public new void Stop()
        {
            m_Lock.WaitOne();

            base.Stop();

            byte[] g723_buffer = new byte[G723_SIZE + 1];

            Array.Clear(g723_buffer, 0, g723_buffer.Length);

            m_Id = Guid.Empty;

            OutVoice(g723_buffer);

            m_Lock.Set();
        }

        private void LKDSPro_ReceiveVoice(uint NumLB, uint LenMsg, ref byte pMsg)
        {
            byte[] Msg = new byte[LenMsg];

            unsafe
            {
                fixed (byte* pBuff = &pMsg)
                {
                    Marshal.Copy((IntPtr)pBuff, Msg, 0, (int)LenMsg);
                }
            }

            G723Pack Pack = new G723Pack((int)LenMsg, Msg);

            if (Pack.GetError() != G723Pack.codError.ceNone)
            {
                if (IsStart)
                {
                    m_Form.PutLog("LU index: " + m_IndexLU.ToString() + " - unavailable or busy");
                    m_Form.Stop();
                }
            }
            else
            {
                byte[] g723_buffer = new byte[G723_SIZE + 1];

                Pack.GetVoice(g723_buffer);

                PutG723(g723_buffer.Length, g723_buffer);
            }
        }

        private void LKDSPro_ReceiveMsg(uint NumLB, uint LenMsg, ref byte pMsg)
        {
            byte[] Msg = new byte[LenMsg];

            unsafe
            {
                fixed (byte* pBuff = &pMsg)
                {
                    Marshal.Copy((IntPtr)pBuff, Msg, 0, (int)LenMsg);
                }
            }

            int MsgLen = Msg[2];
            int MsgTyp = Msg[7];
            int MsgAsk = Msg[9] & 0xF;
            if ( (Msg[8] & 0x20) != 0) MsgAsk |= 0x10;

            if( MsgTyp == 25 || 
                (MsgTyp == 26 && (MsgAsk == 1 || MsgAsk == 2 || MsgAsk == 3 || MsgAsk == 4 || MsgAsk == 7 || MsgAsk == 8)))
            {
                if( (Msg[11] & 0x10) != 0)
                {
                    m_Form.PutLog("Call from LU index: " + NumLB.ToString());
                }
            }
        }


        public override void Put(WavPack aPack)
        {
            byte[] g723_buffer = new byte[G723_SIZE + 1];
            short[] samlpBuf = new short[WAV_SIZE];

            if (m_Direct == sdToDisp)
            {
                m_CountNulOut++;

                if ((m_CountNulOut > 5) && ((m_CountNulOut % 2) != 0)) return;    // skip a package

                Array.Copy(NulOutG723, g723_buffer, g723_buffer.Length);
            }
            else
            {
                Array.Copy(aPack.Data, samlpBuf, WAV_SIZE);

                for (int index = 0; index < G723_FRAME; index++)
                {
                    unsafe
                    {
                        fixed (short* pSamplPart = &samlpBuf[WAV_FRAME_SIZE * index])
                        {
                            fixed (byte* pG723Prt = &g723_buffer[G723_FRAME_SIZE * index])
                            {
                                Coder_G723_1(pSamplPart, pG723Prt);
                            }
                        }
                    }
                }

                g723_buffer[G723_SIZE] = m_Direct;

                m_CountNulOut = 0;
            }

            OutVoice(g723_buffer);
        }

        public void OutVoice(byte[] g723_buffer)
        {
            G723Pack Pack = new G723Pack();

            Pack.SetTyp(G723Pack.typPack.tpMaster);
            Pack.SetVoice(g723_buffer);
            Pack.SetDestPoint(m_Point);
            Pack.SetGUID(m_Id);

            byte[] body = new byte[Pack.Len];
            Array.Copy(Pack.Data, body, Pack.Len);

            m_LKDSPro.SendVoice((uint)m_IndexLU, (uint)Pack.Len, ref body[0]);

            //PutG723(g723_buffer.Length, g723_buffer);       // for voice loop
        }

        public override void Read(out WavPack aPack)
        {
            m_OutQueue.Read(out aPack);
        }
        public override void Break()
        {
            m_OutQueue.Put(new WavPack());
        }
        public void PutG723(int aLenVoice, byte[]aVoice)
        {
            m_Lock.WaitOne();

            short[] Buffer = new short[WAV_SIZE];

            for (int index = 0; index < G723_FRAME; index++)
            {
                if ((aVoice[G723_FRAME_SIZE * index] & 0x3) == 0)
                {
                    unsafe
                    {
                        fixed (short* pSamplPart = &Buffer[WAV_FRAME_SIZE * index])
                        {
                            fixed (byte* pG723Prt = &aVoice[G723_FRAME_SIZE * index])
                            {
                                Decoder_G723_1(pSamplPart, pG723Prt);
                            }
                        }
                    }
                }
                else
                {   // comfortable noise
                    Array.Copy(s_NulWAV, 0, Buffer, WAV_FRAME_SIZE * index, WAV_FRAME_SIZE);
                }
            }

            m_OutQueue.Put(new WavPack(Buffer.Length, Buffer));

            m_Lock.Set();
        }

        public void ChangDirectionVoice()
        {
            m_Direct = m_Direct == sdToLift ? sdToDisp : sdToLift;
        }

        public bool IsListen
        {
            get { return m_Direct == sdToDisp; }
        }

        public const int G723_FRAME = 4;
        public const int G723_FRAME_SIZE = 24;
        public const int G723_SIZE = G723_FRAME_SIZE * G723_FRAME;
        public const int WAV_FRAME_SIZE = 240;
        public const int WAV_SIZE = WAV_FRAME_SIZE * G723_FRAME;

        WavPackQueue m_OutQueue = new WavPackQueue();   // packet queue to audio outpu
        int m_CountNulOut = 0;     // counter to an empty packe
        int m_IndexLU = 0;
        int m_Point = 0;
        Guid m_Id;
        Form1 m_Form;

        byte m_Direct = sdToDisp;	        // direction of sound transmission
        const byte sdToLift = (byte)0x5F;   // direction -> LIFT
        const byte sdToDisp = (byte)0x50;   // direction <- LIFT


        EventWaitHandle m_Lock = new EventWaitHandle(true, EventResetMode.AutoReset);

        LKDSProLib.LKDSMngProClass m_LKDSPro = new LKDSProLib.LKDSMngProClass();


        readonly byte[] s_NulWAV = {
            0x00, 0x00, 0xFC, 0xFF, 0xFE, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xFC, 0xFF, 0xFE, 0xFF, 0xFE, 0xFF, 0xFE, 0xFF, 0xFC, 0xFF, 0xFE, 0xFF, 0xFC, 0xFF,
            0xFE, 0xFF, 0xFE, 0xFF, 0xFE, 0xFF, 0xFE, 0xFF, 0xFE, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0xFE, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x02, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0xFE, 0xFF, 0xFE, 0xFF, 0xFE, 0xFF, 0x00, 0x00, 0x00, 0x00, 0xFE, 0xFF,
            0x00, 0x00, 0x04, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0xFF, 0xFE, 0xFF, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00,
            0x02, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0xFE, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xFE, 0xFF, 0xFE, 0xFF, 0xFE, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFE, 0xFF, 0xFE, 0xFF, 0xFE, 0xFF, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFE, 0xFF, 0xFE, 0xFF, 0xFC, 0xFF, 0xFC, 0xFF, 0xFE, 0xFF,
            0xFE, 0xFF, 0xFC, 0xFF, 0xFE, 0xFF, 0xFE, 0xFF, 0xFE, 0xFF, 0x00, 0x00, 0xFE, 0xFF, 0xFE, 0xFF,
            0x00, 0x00, 0xFE, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFC, 0xFF, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFA, 0xFF, 0xFE, 0xFF, 0x00, 0x00,
            0xFE, 0xFF, 0x00, 0x00, 0xFE, 0xFF, 0x02, 0x00, 0x00, 0x00, 0xFE, 0xFF, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFE, 0xFF, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x04, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0xFE, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00,
            0x00, 0x00, 0xFE, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00,
            0x02, 0x00, 0xFE, 0xFF, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x00, 0x02, 0x00, 0x04, 0x00, 0x02, 0x00, 0x00, 0x00,
            0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0x02, 0x00, 0xFE, 0xFF,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFE, 0xFF};

        readonly byte[] NulOutG723 = {
            0xC2, 0x3A, 0xB6, 0x52, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            sdToDisp};

    };
}
